/**
 * @file SeedService.hh
 * @brief An art service to assist in the distribution of guaranteed unique seeds to all engines within an art job.
 * @author Rob Kutschke (kutschke@fnal.gov)
 * @date   2013/03/14 19:54:49
 * @see SeedService_service.cc
 * 
 */

/******************************************************************************
 *** TODO QA
 *** 
 *** - the original code had copy constructor made private to prevent it from
 ***   being called (now it is deleted); is this a prescription for a service?
 *** 
 ******************************************************************************/

#ifndef SeedService_SeedService_hh
#define SeedService_SeedService_hh

// C/C++ standard libraries
#include <string>
#include <map>
#include <iomanip>

// Some helper classes.
#include "artextensions/SeedService/ArtState.hh"
#include "artextensions/SeedService/SeedMaster.hh"

// From art and its tool chain.
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Services/Optional/RandomNumberGenerator.h"
#include "art/Framework/Services/Registry/ServiceMacros.h"

// Forward declarations
namespace art {
  class ActivityRegistry;
  class ModuleDescription;
  class Run;
  class SubRun;
}

/// artextensions repository namespace
namespace artext {
  
  /**
   * @brief An art service to assist in the distribution of guaranteed unique seeds to all engines within an art job.
   * @see SeedMaster
   *
   * The SeedService acts as an interface between art framework and the
   * SeedMaster class.
   * 
   * 
   * The following instructions presume that the user is familiar with the information on:
   *    http://mu2e.fnal.gov/atwork/computing/Random.shtml#background
   *    http://mu2e.fnal.gov/atwork/computing/Random.shtml#user
   *
   * This class is configured from a FHiCL parameter set:
   *
   *     SeedService : {
   *        policy           : "autoIncrement"  // Required: Other legal value are listed in SEED_SERVICE_POLICIES
   *        baseSeed         : 0                // Required: An integer >= 0.
   *        checkRange       : true             // Optional: legal values true or false; defaults to true
   *        maxUniqueEngines : 20               // Required iff checkRange is true.
   *
   *        verbosity        : 0               // Optional: default=0, no informational printout
   *        endOfJobSummary  : false           // Optional: print list of all managed seeds at end of job.
   *     }
   *
   * The policy parameter tells the service to which algorithm to use and the
   * meaning of the parameters (see below).
   * If the value of the policy parameter is not one of the known policies, the code will
   * throw an exception.
   *
   * If the policy is defined as `autoIncrement` the above FHiCL fragment shows all
   * run time configurable items. Additional parameters for the `preDefinedOffset` policy
   * will be discussed later.
   *
   * Module code requests a seed by making one of the following two calls:
   *
   *     art::ServiceHandle<SeedService>->getSeed();
   *     art::ServiceHandle<SeedService>->getSeed("instanceName");
   *
   * It is the callers responsibility to use the appropriate form. The call to `getSeed` must be
   * done in either the constructor or the `beginRun` method of a module.  When a call is made
   * to `getSeed`, `SeedService` is aware of which module instance is making the call; each
   * module instance is identified by its module label.
   *
   * When `getSeed` is called with a particular module label and instance name, it computes
   * a seed value, saves it and returns it.  If there is a subsequent call to `getSeed` with
   * the same module label and instance name, the service will return the saved value of
   * the seed.  The following text will use the phrase "unique calls to `getSeed`"; two
   * calls with the same module label and instance names are not considered unique.
   *
   * If the policy is set to `autoIncrement`, the seed is set to `baseSeed`+`offset`, where
   * on the first unique call to `getSeed` the `offset` is set to 0; on the second unique
   * call to `getSeed` it is set to 1, and so on.
   *
   * If the policy is set to `linearMapping`, the seed is set to
   * `maxUniqueEngines`*`baseSeed`+`offset`, where on the first unique call to `getSeed`
   * the `offset` is set to 0; on the second unique call to `getSeed` it is set to 1,
   * and so on.
   *
   * If the policy is set to `preDefinedOffset` then, when `getSeed` is called, the service will look
   * into the parameter set to find a defined offset for the specified module label
   * and instance name.  The returned value of the seed will be `baseSeed`+`offset`.
   *
   * The `preDefinedSeed` policy allows to specify the actual seed to be
   * used.  In this case `baseSeed` is not used.  Note that the policy
   * does not impose any constraints on the user-provided set of seeds.
   * In particular, the uniqueness of the seeds is not
   * enforced. Intended for debugging and special tests, use with care.
   *
   * The FHiCL grammar to specify the offsets takes two forms.  If no instance name
   * is given, the offset is given by:
   *
   *     moduleLabel : offset
   *
   * When a module has multiple instances, the offsets are given by:
   *
   *     moduleLabel : {
   *        instanceName1 : offset1
   *        instanceName2 : offset2
   *     }
   * 
   * Note that we do not support a nameless instance and some named ones in the
   * same module. This is just a configuration syntax limitation.
   * `SeedService` does several additional checks, except for the `preDefinedSeed` policy.
   *
   * If one (module label, instance name) has the same seed as another (module label, instance name),
   * the service will throw an exception.
   *
   * If the `checkRange` parameter is set to `true`, and if an offset is generated with a value outside
   * the range ( `0<= offset < maxUniqueEngines-1` ) then the service will throw an exception.
   *
   * It is the responsibility of the user to ensure that the `baseSeed` and `maxUniqueEngines` are chosen
   * it a way that ensures the required level of uniqueness of seeds.  The example grid jobs have
   * a single point of maintenance to achieve this: the user must specify the starting run number
   * for each grid submission.
   */
  class SeedService {
  public:
    
    using seed_t = art::RandomNumberGenerator::seed_t;
    
    SeedService(const fhicl::ParameterSet&, art::ActivityRegistry&);
    
    // Accept compiler written d'tor.  Not copyable or assignable.
    // This class is not copyable or assignable: these methods are not implemented.
    SeedService(SeedService const&) = delete;
    SeedService const& operator=(SeedService const&) = delete;
    SeedService(SeedService&&) = delete;
    SeedService const& operator=(SeedService&&) = delete;
    ~SeedService() = default;
    
    /// Returns the seed value for this module label (instance name).
    seed_t getSeed();
    seed_t getSeed(std::string instanceName);
    
    /// Prints known (EngineId,seed) pairs.
    template<class Stream> void print(Stream&) const;
    
    /// Prints to the framework Info logger
    void print() const;
    
    // Call backs that will be called by art.
    void preModuleConstruction (art::ModuleDescription const& md);
    void postModuleConstruction(art::ModuleDescription const& md);
    void preModuleBeginRun     (art::ModuleDescription const& md);
    void postModuleBeginRun    (art::ModuleDescription const& md);
    void postEndJob();
    
  private:
    using SeedMaster_t = SeedMaster<seed_t>;
    using EngineId = SeedMaster_t::EngineId;
    
    
    /// Class managing the seeds
    SeedMaster_t seeds;
    
    /// Helper to track state of art. It is legal to call getSeed only in certain states.
    SeedServiceHelper::ArtState state;
    
    /// Control the level of information messages.
    int verbosity = 0;
    bool bPrintEndOfJobSummary = false; ///< print a summary at the end of job
    
    // Main logic for computing and validating a seed.
    seed_t getSeed(EngineId const&);
    
    // Helper functions for all policies
    void ensureValidState() const;
    
  }; // class SeedService
  
  
  
  //================================================================
  template<class Stream>
  inline void SeedService::print(Stream& log) const {
    seeds.print(log);
  } // print(Stream)
  
} // namespace artext

DECLARE_ART_SERVICE(artext::SeedService, LEGACY)

#endif /* SeedService_SeedService_hh */
